#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include <SDL/SDL.h>
#include <png.h>

#include "sauna.h"

/*** Load PNG ***/
/*** PNG loading code is based on IMG_png from SDL_image ***/

static void pngReadData(png_structp ctx, png_bytep area, png_size_t size)
{
	SDL_RWops *src;

	src = (SDL_RWops *) png_get_io_ptr(ctx);
	SDL_RWread(src, area, size, 1);
}

SDL_Surface *loadPng(const char *filename)
{
	SDL_RWops *rw;
	SDL_Surface *surface;

	rw = SDL_RWFromFile(filename, "rb");
	if(!rw)
	{
		fprintf(stderr, "loadPng(\"%s\"): %s\n", filename, SDL_GetError());
		return NULL;
	}

	surface = loadPngRW(rw);
	SDL_RWclose(rw);

	return surface;
}

SDL_Surface *loadPngRW(SDL_RWops * rw)
{
	SDL_Surface *surface;
	SDL_Palette *palette;
	png_structp pngPtr;
	png_infop pngInfoPtr;
	png_uint_32 width, height;
	int bitDepth, colorType, interlaceType;
	Uint32 Rmask, Gmask, Bmask, Amask;
	unsigned char header[4];
	volatile int ckey = -1;
	png_color_16 *transv;
	png_bytep *rowPointers;
	int row, i;

	SDL_RWread(rw, header, 1, 4);
	if(png_sig_cmp(header, 0, 4))
	{
		fprintf(stderr, "Error: loadPngRw(): Not a PNG file !\n");
		return NULL;
	}
	SDL_RWseek(rw, -4, SEEK_CUR);

	pngPtr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if(!pngPtr)
		return NULL;
	pngInfoPtr = png_create_info_struct(pngPtr);

	if(!pngInfoPtr)
	{
		png_destroy_read_struct(&pngPtr, NULL, NULL);
		return NULL;
	}

	if(setjmp(pngPtr->jmpbuf))
	{
		perror("setjmp");
		png_destroy_read_struct(&pngPtr, &pngInfoPtr, NULL);
		return NULL;
	}

	png_set_read_fn(pngPtr, rw, pngReadData);

	png_read_info(pngPtr, pngInfoPtr);
	png_get_IHDR(pngPtr, pngInfoPtr, &width, &height, &bitDepth, &colorType,
							 &interlaceType, NULL, NULL);

	png_set_strip_16(pngPtr);

	png_set_packing(pngPtr);

	if(colorType == PNG_COLOR_TYPE_GRAY)
		png_set_expand(pngPtr);

	if(png_get_valid(pngPtr, pngInfoPtr, PNG_INFO_tRNS))
	{
		int num_trans;
		Uint8 *trans;
		png_get_tRNS(pngPtr, pngInfoPtr, &trans, &num_trans, &transv);
		if(colorType == PNG_COLOR_TYPE_PALETTE)
		{
			int i, t = 1;
			for(i = 0; i < num_trans; i++)
				if(trans[i] == 0)
				{
					if(t >= 0)
						break;
					t = i;
				} else if(trans[i] != 255)
					break;
			if(i == num_trans)
			{
				ckey = t;
			} else
			{
				png_set_expand(pngPtr);
			}
		} else
			ckey = 0;
	}
	if(colorType == PNG_COLOR_TYPE_GRAY_ALPHA)
		png_set_gray_to_rgb(pngPtr);
	png_read_update_info(pngPtr, pngInfoPtr);
	png_get_IHDR(pngPtr, pngInfoPtr, &width, &height, &bitDepth, &colorType,
							 &interlaceType, NULL, NULL);
	/* init SDL surface */
	if(colorType != PNG_COLOR_TYPE_PALETTE)
	{
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		Rmask = 0x000000FF;
		Gmask = 0x0000FF00;
		Bmask = 0x00FF0000;
		Amask = (pngInfoPtr->channels == 4) ? 0xFF000000 : 0;
#else
		int s = (pngInfoPtr->channels == 4) ? 0 : 8;
		Rmask = 0xFF000000 >> s;
		Gmask = 0x00FF0000 >> s;
		Bmask = 0x0000FF00 >> s;
		Amask = 0x000000FF >> s;
#endif
	} else
	{
		Rmask = 0;
		Gmask = 0;
		Bmask = 0;
		Amask = 0;
	}

	surface =
		SDL_AllocSurface(SDL_SWSURFACE, width, height,
										 bitDepth * pngInfoPtr->channels, Rmask, Gmask, Bmask,
										 Amask);
	if(!surface)
	{
		fprintf(stderr, "LoadPngRW(): Out of memory !\n");
		png_destroy_read_struct(&pngPtr, &pngInfoPtr, NULL);
		return NULL;
	}
	if(ckey != -1)
	{
		if(colorType != PNG_COLOR_TYPE_PALETTE)
			ckey =
				SDL_MapRGB(surface->format, (Uint8) transv->red,
									 (Uint8) transv->green, (Uint8) transv->blue);
		SDL_SetColorKey(surface, SDL_SRCCOLORKEY, ckey);
	}
	rowPointers = (png_bytep *) malloc(sizeof(png_bytep) * height);
	if(!rowPointers)
	{
		perror("malloc");
		png_destroy_read_struct(&pngPtr, &pngInfoPtr, NULL);
		SDL_FreeSurface(surface);
		return NULL;
	}

	for(row = 0; (unsigned int)row < height; row++)
	{
		rowPointers[row] =
			(png_bytep) (Uint8 *) surface->pixels + row * surface->pitch;
	}
	png_read_image(pngPtr, rowPointers);
	png_read_end(pngPtr, pngInfoPtr);

	palette = surface->format->palette;
	if(palette)
	{
		if(colorType == PNG_COLOR_TYPE_GRAY)
		{
			palette->ncolors = 256;
			for(i = 0; i < 256; i++)
			{
				palette->colors[i].r = i;
				palette->colors[i].g = i;
				palette->colors[i].b = i;
			}
		} else if(pngInfoPtr->num_palette > 0)
		{
			palette->ncolors = pngInfoPtr->num_palette;
			for(i = 0; i < pngInfoPtr->num_palette; i++)
			{
				palette->colors[i].b = pngInfoPtr->palette[i].blue;
				palette->colors[i].g = pngInfoPtr->palette[i].green;
				palette->colors[i].r = pngInfoPtr->palette[i].red;

			}
		}
	}
	png_destroy_read_struct(&pngPtr, &pngInfoPtr, NULL);
	free(rowPointers);
	return surface;

}

/*** Save a surface as a PNG image ***/
/* Note ! Palette based images are not supported */
int savePng(SDL_Surface * surface, const char *filename)
{
	FILE *fp;
	png_structp pngPtr;
	png_infop pngInfoPtr;
	png_bytep *rowPointers;
	int i;

	fp = fopen(filename, "wb");
	if(!fp)
	{
		perror("fopen");
		return 1;
	}

	pngPtr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	pngInfoPtr = png_create_info_struct(pngPtr);
	png_init_io(pngPtr, fp);

	pngInfoPtr->width = surface->w;
	pngInfoPtr->height = surface->h;
	pngInfoPtr->rowbytes = surface->pitch;
	pngInfoPtr->bit_depth = 8;
	pngInfoPtr->interlace_type = 0;
	pngInfoPtr->num_palette = 0;
	pngInfoPtr->valid = 0;

	pngInfoPtr->sig_bit.red = 8;
	pngInfoPtr->sig_bit.green = 8;
	pngInfoPtr->sig_bit.blue = 8;
	pngInfoPtr->sig_bit.alpha = 8;

	pngInfoPtr->color_type = PNG_COLOR_TYPE_RGB_ALPHA;

	png_write_info(pngPtr, pngInfoPtr);

	rowPointers = (png_bytep *) malloc(sizeof(png_bytep) * pngInfoPtr->height);

	for(i = 0; (unsigned int)i < pngInfoPtr->height; i++)
	{
		rowPointers[i] = (Uint8*)surface->pixels + i * surface->pitch;
	}

	png_write_image(pngPtr, rowPointers);
	png_write_end(pngPtr, pngInfoPtr);
	free(rowPointers);
	png_destroy_write_struct(&pngPtr, &pngInfoPtr);
	return 0;
}

/* Font routines */
SDL_Surface *font;

void drawString(char *str, int x, int y)
{
	SDL_Rect srect, drect;
	unsigned int i, c;
	
	srect.w = drect.w = font->w / 16;
	srect.h = drect.h = font->h / 16;
	drect.x = x;
	drect.y = y;
	
	for(i = 0; i < strlen(str); i++)
	{
		c = (unsigned int)str[i];
		if(c != '\n')
		{
			srect.x = (c % 16) * srect.w;
			srect.y = (c / 16) * srect.h;
			SDL_BlitSurface(font, &srect, screen, &drect);
			drect.x += drect.w;
		} else
		{
			drect.x = x;
			drect.y += drect.h;
		}
	}
}

int gprintf(int x, int y,const char *str, ...)
{
	char buf[512];
	int len;

	va_list ap;
	va_start(ap, str);

	len = vsnprintf(buf, 512, str, ap);

	va_end(ap);
	drawString(buf, x, y);

	return len;
}

static Uint32 getpixel(SDL_Surface *surface, int x, int y)
{
	int bpp = surface->format->BytesPerPixel;
	/* Here p is the address to the pixel we want to retrieve */
	Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

	switch(bpp) {
	case 1:
		return *p;
	case 2:
		return *(Uint16 *)p;
	case 3:
	if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
		return p[0] << 16 | p[1] << 8 | p[2];
	else
		return p[0] | p[1] << 8 | p[2] << 16;
	case 4:
		return *(Uint32 *)p;
	default:
		return 0;       /* shouldn't happen, but avoids warnings */
	}
}

static void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel)
{
	int bpp = surface->format->BytesPerPixel;
	/* Here p is the address to the pixel we want to set */
	Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

	switch(bpp) {
		case 1:
			*p = pixel;
			break;
		case 2:
			*(Uint16 *)p = pixel;
			break;
		case 3:
			if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
				p[0] = (pixel >> 16) & 0xff;
				p[1] = (pixel >> 8) & 0xff;
				p[2] = pixel & 0xff;
			} else {
				p[0] = pixel & 0xff;
				p[1] = (pixel >> 8) & 0xff;
				p[2] = (pixel >> 16) & 0xff;
			}
			break;
		case 4:
			*(Uint32 *)p = pixel;
			break;
	}
}

void surfaceGreyscale(SDL_Surface *surface) {
	int x,y;
	if ( SDL_MUSTLOCK(surface) ) {
		if ( SDL_LockSurface(surface) < 0 ) {
			fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
			return;
		}
	}

	for(x=0;x<surface->w;x++)
		for(y=0;y<surface->h;y++) {
			Uint32 col;
			Uint8 r,g,b;
			col=getpixel(surface,x,y);
			SDL_GetRGB(col,surface->format,&r,&g,&b);
			r=(r+g+b)/3;
			col=SDL_MapRGB(surface->format,r,r,r);
			putpixel(surface,x,y,col);
		}
	if ( SDL_MUSTLOCK(screen) ) {
		SDL_UnlockSurface(screen);
	}
}

char *inputBox(const char *prompt,int width,int maxlen) {
	char *buffer;
	SDL_Rect rect;
	SDL_Event event;
	Uint32 white,black;
	int len,l;

	buffer=malloc(maxlen+1);
	rect.w=width;
	rect.h=100;
	rect.x=screen->w/2-rect.w/2;
	rect.y=screen->h/2-rect.h/2;
	white=SDL_MapRGB(screen->format,255,255,255);
	black=SDL_MapRGB(screen->format,0,0,0);
	SDL_FillRect(screen,&rect,white);
	rect.x+=2; rect.w-=4;
	rect.y+=2; rect.h-=4;
	SDL_FillRect(screen,&rect,black);
	
	len=strlen(prompt)*8;
	gprintf(screen->w/2-len/2,screen->h/2-16,"%s",prompt);
	buffer[0]='\0'; l=0;
	SDL_Flip(screen);

	SDL_EnableUNICODE(1);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY,SDL_DEFAULT_REPEAT_INTERVAL);
	rect.h=8;
	rect.y=screen->h/2;

	while(1) {
		if(!SDL_WaitEvent(&event)) {
      		fprintf(stderr,"Error occured while waiting for an event: %s\n",SDL_GetError());
			SDL_EnableUNICODE(0);
			SDL_EnableKeyRepeat(0,SDL_DEFAULT_REPEAT_INTERVAL);
			free(buffer);
			return NULL;
    	}
		if(event.type==SDL_QUIT) {
				SDL_EnableUNICODE(0);
				SDL_EnableKeyRepeat(0,SDL_DEFAULT_REPEAT_INTERVAL);
				free(buffer);
				return NULL;
		} else if(event.type==SDL_KEYDOWN) {
			if(event.key.keysym.sym==SDLK_RETURN || event.key.keysym.sym==SDLK_KP_ENTER) {
				break;
			} else if(event.key.keysym.sym==SDLK_BACKSPACE) {
				if(l==0) continue;
				rect.w=l*8;
				rect.x=screen->w/2-rect.w/2;
				SDL_FillRect(screen,&rect,black);
				buffer[--l]='\0';
				gprintf(rect.x+4,rect.y,"%s",buffer);
				SDL_UpdateRect(screen,rect.x,rect.y,rect.w,rect.h);
			} else if(isalnum(event.key.keysym.unicode) || ispunct(event.key.keysym.unicode) || event.key.keysym.unicode==' ') {
				if(l>=maxlen) continue;
				buffer[l++]=event.key.keysym.unicode;
				buffer[l]='\0';
				rect.w=l*8;
				rect.x=screen->w/2-rect.w/2;
				SDL_FillRect(screen,&rect,black);
				gprintf(rect.x,rect.y,"%s",buffer);
				SDL_UpdateRect(screen,rect.x,rect.y,rect.w,rect.h);
			}
		}
	}
	SDL_EnableUNICODE(0);
	SDL_EnableKeyRepeat(0,SDL_DEFAULT_REPEAT_INTERVAL);
	return buffer;
}

